"""Data StarLine API."""
import logging
from typing import Dict, List, Callable, Optional, Any
from .base_api import BaseApi
from .device import StarlineDevice

_LOGGER = logging.getLogger(__name__)


class StarlineApi(BaseApi):
    """Data StarLine API class."""

    def __init__(self, user_id: str, slnet_token: str):
        """Constructor."""
        super().__init__()
        self._user_id = user_id
        self._slnet_token = slnet_token
        self._devices: Dict[str, StarlineDevice] = {}
        self._update_listeners: List[Callable] = []

    def _call_listeners(self) -> None:
        """Call listeners for update notifications."""
        for listener in self._update_listeners:
            listener()

    def add_update_listener(self, listener: Callable) -> None:
        """Add a listener for update notifications."""
        self._update_listeners.append(listener)

    async def update(self, unused=None) -> None:
        """Update StarLine data."""
        devices = await self.get_user_info()

        for device_data in devices:
            device_id = str(device_data["device_id"])
            if device_id not in self._devices:
                self._devices[device_id] = StarlineDevice()
            self._devices[device_id].update(device_data)

        self._call_listeners()

    @property
    def devices(self):
        """Devices list."""
        return self._devices

    async def get_user_info(self) -> Optional[List[Dict[str, Any]]]:
        """Get user information."""

        # TODO: handle {'code': 500, 'codestring': 'Bad user id'}
        url = "https://developer.starline.ru/json/v2/user/{}/user_info".format(
            self._user_id
        )
        headers = {"Cookie": "slnet=" + self._slnet_token}
        response = await self.get(url, headers=headers)

        code = int(response["code"])
        if code == 200:
            return response["devices"] + response["shared_devices"]
        return None

    async def set_car_state(self, device_id: str, name: str, state: bool):
        """Set car state information."""
        _LOGGER.debug("Setting car %s state: %s=%d", device_id, name, state)
        url = "https://developer.starline.ru/json/v1/device/{}/set_param".format(
            device_id
        )
        data = {"type": name, name: 1 if state else 0}
        headers = {"Cookie": "slnet=" + self._slnet_token}
        response = await self.post(url, json=data, headers=headers)

        code = int(response["code"])
        if code == 200:
            self._devices[device_id].update_car_state(response)
            self._call_listeners()
            return response
        return None



"""Base StarLine API."""
import aiohttp
import logging
from .const import DEFAULT_CONNECT_TIMEOUT, DEFAULT_TOTAL_TIMEOUT, DEFAULT_ENCODING, GET, POST

_LOGGER = logging.getLogger(__name__)


class BaseApi:
    """Base StarLine API class."""

    def __init__(self):
        """Constructor."""
        self._session: aiohttp.ClientSession = aiohttp.ClientSession()
        self._total_timeout: int = DEFAULT_TOTAL_TIMEOUT
        self._connect_timeout: int = DEFAULT_CONNECT_TIMEOUT
        self._encoding: str = DEFAULT_ENCODING

    def set_timeout(self, total_timeout: int, connect_timeout: int = None) -> None:
        self._total_timeout = total_timeout
        if connect_timeout is not None:
            self._connect_timeout = connect_timeout

    def set_encoding(self, encoding: str) -> None:
        self._encoding = encoding

    async def _request(
        self,
        method: str,
        url: str,
        params: dict = None,
        data: dict = None,
        json: dict = None,
        headers: dict = None,
    ) -> aiohttp.ClientResponse:
        """Make request."""

        response = await self._session.request(
            method,
            url,
            params=params,
            data=data,
            json=json,
            headers=headers,
            timeout=aiohttp.ClientTimeout(
                total=self._total_timeout,
                connect=self._connect_timeout
            ),
        )
        response.raise_for_status()
        response.encoding = self._encoding

        _LOGGER.debug("StarlineApi {} request: {}".format(method, url))
        _LOGGER.debug("  Payload: {}".format(params))
        _LOGGER.debug("  Data: {}".format(data))
        _LOGGER.debug("  JSON: {}".format(json))
        _LOGGER.debug("  Headers: {}".format(headers))
        _LOGGER.debug("  Response: {}".format(response))

        # TODO: Handle Exceptions
        return response

    async def get(self, url: str, params: dict = None, headers: dict = None) -> dict:
        """Make GET request."""

        response = await self._request(GET, url, params=params, headers=headers)
        data = await response.json(content_type=None)
        _LOGGER.debug("  Data: {}".format(data))
        return data

    async def post(
        self,
        url: str,
        params: dict = None,
        data: dict = None,
        json: dict = None,
        headers: dict = None,
    ) -> dict:
        """Make POST request."""

        response = await self._request(
            POST, url, params=params, data=data, json=json, headers=headers
        )
        data = await response.json(content_type=None)
        _LOGGER.debug("  Data: {}".format(data))
        return data



"""Auth StarLine API."""
import hashlib
import logging
import re
from datetime import datetime
from .base_api import BaseApi, POST

_LOGGER = logging.getLogger(__name__)


class StarlineAuth(BaseApi):
    """Auth StarLine API class."""

    async def get_app_code(self, app_id: str, app_secret: str) -> str:
        """Get application code for getting application token."""

        url = "https://id.starline.ru/apiV3/application/getCode/"
        payload = {
            "appId": app_id,
            "secret": hashlib.md5(app_secret.encode(self._encoding)).hexdigest(),
        }
        response = await self.get(url, params=payload)

        if int(response["state"]) == 1:
            app_code = response["desc"]["code"]
            _LOGGER.debug("Application code: {}".format(app_code))
            return app_code
        raise Exception(response)

    async def get_app_token(self, app_id: str, app_secret: str, app_code: str) -> str:
        """Get application token for authentication."""

        url = "https://id.starline.ru/apiV3/application/getToken/"
        payload = {
            "appId": app_id,
            "secret": hashlib.md5((app_secret + app_code).encode(self._encoding)).hexdigest(),
        }
        response = await self.get(url, params=payload)

        if int(response["state"]) == 1:
            app_token = response["desc"]["token"]
            _LOGGER.debug("Application token: {}".format(app_token))
            return app_token
        raise Exception(response)

    async def get_slid_user_token(
        self,
        app_token: str,
        user_login: str,
        user_password: str,
        sms_code: str = None,
        captcha_sid: str = None,
        captcha_code: str = None,
    ) -> (bool, dict):
        """Authenticate user by login, password and application token."""

        url = "https://id.starline.ru/apiV3/user/login/"
        payload = {"token": app_token}
        data = {
            "login": user_login,
            "pass": hashlib.sha1(user_password.encode(self._encoding)).hexdigest(),
        }
        if sms_code is not None:
            data["smsCode"] = sms_code
        if (captcha_sid is not None) and (captcha_code is not None):
            data["captchaSid"] = captcha_sid
            data["captchaCode"] = captcha_code
        response = await self.post(url, params=payload, data=data)

        state = int(response["state"])
        if (
            (state == 1)
            or (state == 2)
            or (state == 0 and "captchaSid" in response["desc"])
            or (state == 0 and "phone" in response["desc"])
        ):
            return state, response["desc"]
        raise Exception(response)

    async def get_user_id(self, slid_token: str) -> (str, float, str):
        """Authenticate user by StarLineID token."""
        # TODO: check response code
        # TODO: Продливать куку

        url = "https://developer.starline.ru/json/v2/auth.slid"
        data = {"slid_token": slid_token}
        response = await self._request(POST, url, json=data)
        json = await response.json(content_type=None)

        # Read cookie from headers because of bug https://gitlab.com/starline/openapi/issues/3
        cookie_header = response.headers.get("Set-Cookie")
        slnet = re.search("slnet=([^;]+);", cookie_header)
        expires = re.search("expires=([^;]+);", cookie_header)
        if slnet is None:
            raise Exception(response)

        slnet_token = slnet.group(1)
        expires_time = None

        if expires is not None:
            try:
                expires_time = datetime.strptime(expires.group(1), '%A, %d-%b-%y %H:%M:%S %Z').timestamp()
            except:
                pass

        _LOGGER.debug("SLnet token: {}".format(slnet_token))
        return slnet_token, expires_time, json["user_id"]
